package org.khelekore.jsizer;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/** A graph builder that uses full class names to build the graph.
 */
public class HierarchyGraphBuilder implements GraphBuilder {
    private ColorHandler colorHandler;

    public HierarchyGraphBuilder (ColorHandler colorHandler) {
	this.colorHandler = colorHandler;
    }

    public boolean hasFullName () {
	return true;
    }

    public List<TreeNode> buildGraph (List<ClassSize> classSizes) {
	Map<String, ClassSize> name2ClassSize = 
	    new HashMap<String, ClassSize> ();
	Map<String, List<ClassSize>> superClass2Children = 
	    new HashMap<String, List<ClassSize>> ();

	// gather up existing classes and build child lists for super classes
	for (ClassSize cs : classSizes) {
	    String cn = getName (cs.getClassName ());
	    String sn = getName (cs.getSuperName ());
	    name2ClassSize.put (cn, cs);
	    
	    List<ClassSize> children = superClass2Children.get (sn);
	    if (children == null) {
		children = new ArrayList<ClassSize> ();
		superClass2Children.put (sn, children);
	    }
	    children.add (cs);
	}
	
	// use the above information to build the tree of Entry nodes
	// TODO: skip Entry, go directly to TreeNode.
	List<Entry> ls = new ArrayList<Entry> ();
	Set<String> handled = new HashSet<String> ();
	for (ClassSize cs : classSizes) {
	    String superName = getName (cs.getSuperName ());
	    if (!name2ClassSize.containsKey (superName) && 
		!handled.contains (superName)) {
		Entry foo = new Entry (superName, null, 
				       collectChildren (superClass2Children, 
							superName));
		ls.add (foo);
		handled.add (superName);
	    }
	}

	Collections.sort (ls, new EntryNameSorter ());
	
	// convert Entry into TreeNode
	List<TreeNode> ret = new ArrayList<TreeNode> ();
	for (Entry entry : ls) {
	    TreeNode tn = makeTreeNode (entry);
	    ret.add (tn);
	}
	return ret;
    }

    private static class EntryNameSorter implements Comparator<Entry> {
	public int compare (Entry e1, Entry e2) {
	    String n1 = e1.name;
	    String n2 = e2.name;
	    return n1.compareTo (n2);
	}
    }

    private String getName (String name) {
	return name.replaceAll ("[$/]", ".");
    }

    private List<Entry> collectChildren (Map<String, List<ClassSize>> superClass2Children, 
					 String name) {
	List<ClassSize> children = superClass2Children.get (name);
	if (children == null)
	    return null;
	List<Entry> entries = new ArrayList<Entry> (children.size ());
	for (ClassSize cs : children) {
	    String cname = cs.getClassName ();
	    entries.add (new Entry (cname, cs, 
				    collectChildren (superClass2Children, 
						     cname)));
	}
	Collections.sort (entries, new EntryNameSorter ());
	return entries;
    }

    private TreeNode getTreeNode (ClassSize cs, String name) {
	if (cs != null) {
	    Color color = colorHandler.getColor (cs);
	    TreeNode tn = new TreeNode (name, 
					cs.getNumFields (), 
					cs.getNumMethods (), 
					cs.getNumMethodLines (), 
					color);
	    return tn;
	} 
	return new TreeNode (name);
    }

    private TreeNode makeTreeNode (Entry entry) {
	TreeNode tn = getTreeNode (entry.cs, entry.name);
	if (entry.children != null) {
	    for (Entry e : entry.children) {
		TreeNode cn = makeTreeNode (e);
		tn.addChild (cn);
	    }
	}
	return tn;
    }
    
    private static class Entry {
	public String name;
	public ClassSize cs;
	public List<Entry> children;

	public Entry (String name, ClassSize cs, List<Entry> children) {
	    this.name = name;
	    this.cs = cs;
	    this.children = children;
	}
    }
}
